package homology;

import topology.*;
import homology.f2homology.*;

/**
 * This class provides methods for computing the homology of finite topological
 * spaces.
 * @param <K> space type
 * 
 * @author Patrick Erik Bradley
 * @version 1.0 --- May 30, 2014
 */
@SuppressWarnings("unchecked")
public class Betti<K extends Comparable<K>> {
    /**
     * the space whose Betti numbers are to be computed
     * @param <K> the space type
     */
    private FiniteSpace<K> space = new FiniteSpace<K>();
    /**
     * the core of the space whose Betti numbers are to be computed
     * @param <K> space type
     */
    private FiniteSpace<K> core = new FiniteSpace<K>();
    private int dimension = -1; 

    /**
     * Creates a new Betti object.
     */
    public Betti() {
    }

    /**
     * Creates an instance of the Betti class 
     * for a given finite topological space. A point reduction of the space
     * is computed by iteratively removing weak points of its Kolmogorov
     * quotient. This is the core of this space. It can possibly be 
     * further reduced without altering the Betti numbers by future methods. 
     * @param <K> space type
     * @param space a finite space
     */
    public Betti(FiniteSpace<K> space) {
	this.space = new FiniteSpace(space);
	//core = new FiniteSpace(space.T0Space());
	dimension = space.dimension();
	FiniteSpace<K> aSpace 
	    = new FiniteSpace(PointReductions.removeWeakPoints(space));
	core = new FiniteSpace();
	core.addAllPoint(aSpace.getPointSet());
	core.addAllPair(aSpace.getRelation().getPairs());
    }

    /**
     * Returns this space.
     * @param <K> space type
     * @return space
     */
    public FiniteSpace<K> getSpace() {
	return space;
    }

    /**
     * Returns the core of this space.
     * @param <K> space type
     * @return core
     */
    public FiniteSpace<K> getCore() {
	return core;
    }

    /**
     * Returns the order complex of the core of this space.
     * @param <K> space type
     * @return order complex
     */
    public OrderComplex<K> orderComplex() {
	OrderComplex<K> orderComplex = new OrderComplex(core);
	return orderComplex;
    }

    /**
     * Returns the Betti numbers over the finite field with two elements.
     * @return array of Betti numbers over F2
     */
    public int[] F2BettiNumbers() {

	int[] BettiNumbers = new int[dimension + 1];
	OrderComplex<K> orderComplex = orderComplex();
	F2ChainComplex<Integer> chainComplex = orderComplex.F2ChainComplex();
	int[] betti = chainComplex.BettiNumbers();
	for (int i = 0; i < betti.length; i++) {
	    BettiNumbers[i] = betti[i];
	}
	for (int i = betti.length; i < BettiNumbers.length; i++) {
	    BettiNumbers[i] = 0;
	}
	return BettiNumbers;
    }
    
    /**
     * Returns the homological dimension over 
     * the finite field with two elements.
     * @return homological dimension over F2
     */
    public int F2HomologicalDimension() {
	int homDim = -1;
	int[] BettiNumbers = F2BettiNumbers();
	int N = BettiNumbers.length;
	for (int i = 0; i < N; i++) {
	    if (BettiNumbers[i] > 0) {
		homDim = i;
	    }
	}
	return homDim;
    }
}
