package homology;

import java.util.ArrayList;
import java.util.TreeMap;
import java.util.Map;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.LinkedList;
import java.util.ListIterator;
import topology.*;
import homology.f2homology.*;

/**
 * This class represents the order complex associated to a 
 * finite topological space.
 * @param <K> index type
 * 
 * @author Patrick Erik Bradley
 * @version 1.0 --- May 13, 2014
 */
@SuppressWarnings("unchecked")
public class OrderComplex<K extends Comparable<K>> {
    /**
     * the set of maximal chains in the space whose order complex 
     * is to be computed
     * @param <K> space type
     */
    private TreeSet<List<K>> maximalChains = new TreeSet<List<K>>(new ListComparator());

    /**
     * Constructs a new empty order complex.
     */
    public OrderComplex() {
    }

    /** 
     * Constructs an order complex from a finite topological space. 
     * @param <K> index type
     * @param space finite topological space 
     */
    public OrderComplex(FiniteSpace<K> space) {
	FiniteSpace aSpace = new FiniteSpace();
	aSpace.addAllPoint(space.T0Space().getPointSet());
	aSpace.addAllPair(space.T0Space().getRelation().getPairs());
	//	new FiniteSpace(space.T0Space());
	Set<K> vertices = aSpace.KVertices();
	Set<K> opens = aSpace.KOpenPoints();
	Iterator<K> vIter = vertices.iterator();
	while (vIter.hasNext()) {
	    K v = vIter.next();
	    Iterator<K> oIter = opens.iterator();
	    while (oIter.hasNext()) {
		K o = oIter.next();
		maximalChains.addAll(space.allPaths(o,v));
	    }
	}
    }

    /**
     * Returns the dimension of this order complex.
     * @return dimension
     */
    public int dimension() {
	int dimension = -1;
	if (!maximalChains.isEmpty()) {
	    List<K> maxChain = maximalChains.first();
	    Iterator<List<K>> iter = maximalChains.iterator();
	    while (iter.hasNext()) {
		List<K> chain = iter.next();
		if (chain.size() > maxChain.size()) {
		    maxChain = chain;
		}
	    }
	    dimension = maxChain.size() - 1;
	}
	return dimension;
    }

    /**
     * Returns the set of maximal chains as a set of linked lists.
     * @param <K> index type
     * @return maximalChains
     */
    public TreeSet<List<K>> getMaximalChains() {
	return maximalChains;
    }

    /**
     * Returns the set of maximal simplices as a set of sets.
     * @param <K> index type
     * @return maximalSimplices
     */
    public Set<Set<K>> maximalSimplices() {
	Set<Set<K>> maximalSimplices = new TreeSet<Set<K>>(new TreeSetComparator());
	Iterator<List<K>> iter = maximalChains.iterator();
	while (iter.hasNext()) {
	    List<K> s = iter.next();
	    Set<K> simplex = new TreeSet<K>();
	    Iterator<K> sIter = s.iterator();
	    while (sIter.hasNext()) {
		simplex.add(sIter.next());
	    }
	    maximalSimplices.add(simplex);
	}
	return maximalSimplices;
    }

    /**
     * Returns the set of all simplices as a set of sets.
     * @param <K> index type
     * @return simplices
     */
    public Set<Set<K>> simplices() {
	Set<Set<K>> simplices = new TreeSet<Set<K>>(new TreeSetComparator());
	for (Set<K> s : maximalSimplices()) {
	    simplices.addAll(PowerSet(s));
	}
	Iterator<Set<K>> iter = simplices.iterator();
	while(iter.hasNext()) {
	    Set<K> s = iter.next();
	    if (s.isEmpty()) {
		iter.remove();
	    }
	}
	return simplices;
    }

    /**
     * Returns a bijective map between the set of all simplices 
     * and a set of integers.
     * @return indices
     */
    public TreeMap<Set<K>,Integer> indexMap() {
	TreeMap<Set<K>,Integer> indexMap 
	    = new TreeMap<Set<K>,Integer>(new TreeSetComparator());
	int index = -1;
	Iterator<Set<K>> iter = simplices().iterator();
	while (iter.hasNext()) {
	    Set<K> simplex = iter.next();
	    index = index + 1;
	    indexMap.put(simplex,index);
	}
	return indexMap;
    }
    
    public TreeMap<Integer,Set<K>> inverseMap() {
	TreeMap<Integer,Set<K>> inverseMap = new TreeMap<Integer,Set<K>>();
	int index = -1;
	Iterator<Set<K>> iter = simplices().iterator();
	while (iter.hasNext()) {
	    Set<K> simplex = iter.next();
	    index = index + 1;
	    inverseMap.put(index,simplex);
	}
	return inverseMap;
    }
    
    /**
     * Returns the power set of a set.
     * @param <T> index type
     * @param set
     * @return power set
     */
    private static <T> Set<Set<T>> PowerSet(Set<T> set) {
	Set<Set<T>> powerSet = new TreeSet<Set<T>>(new TreeSetComparator());
	if (set.isEmpty()) {
	    powerSet.add(new TreeSet<T>());
	} else {
	    List<T> list = new ArrayList<T>(set);
	    T head = list.get(0);
	    Set<T> rest = new TreeSet<T>(list.subList(1, list.size())); 
	    for (Set<T> s : PowerSet(rest)) {
		Set<T> newSet = new TreeSet<T>();
		newSet.add(head);
		newSet.addAll(s);
		powerSet.add(newSet);
		powerSet.add(s);
	    }	
	}
	return powerSet;
    }

    /**
     * Returns a list of sets of simplex indices. The n-th entry is the set of 
     * indices for simplices of dimension n.
     * @return list of simplex sets
     */
    public List<TreeSet<Integer>> listOfSimplexSets() {
	List<TreeSet<Integer>> listOfSimplexSets 
	    = new ArrayList();
	int n = dimension();
	for (int i = 0; i < n + 1; i++) {
	    TreeSet<Integer> simplexSet = new TreeSet();
	    listOfSimplexSets.add(simplexSet);
	}
	Iterator<Set<K>> simpIter = simplices().iterator();
	while (simpIter.hasNext()) {
	    Set<K> simplex = simpIter.next();
	    int s = simplex.size();
	    int index = indexMap().get(simplex);
	    listOfSimplexSets.get(s-1).add(index);
	}
	return listOfSimplexSets;
    }
 
    /**
     * Returns an array of all simplices. The n-th entry is the set of 
     * simplices of dimension n.
     * @return array of simplices
     */
    public Set<Set<K>>[] arrayOfSimplices() {
	
    	Set<Set<K>>[] arrayOfSimplices 
	    = (Set<Set<K>>[]) new Object[dimension()];
	Iterator<Set<K>> iter = simplices().iterator();
	while (iter.hasNext()) {
	    Set<K> simplex = iter.next();
	    int s = simplex.size();
	    arrayOfSimplices[s-1].add(simplex);
	}
	return arrayOfSimplices;
    }
    
    /**
     * Returns true if set A is a subset of set B.
     * @param A set A
     * @param B set B
     * @return isSubSet
     */
    private <T> boolean isSubSet(Set<T> A, Set<T> B) {
	boolean isSubSet = B.containsAll(A);
	return isSubSet;
    }

    /**
     * Returns the chain complex over the field F2 associated to 
     * this order complex.
     * @param <K> index type
     * @return chain complex
     */
    public F2ChainComplex<Integer> F2ChainComplex() {
	
	F2ChainComplex<Integer> chainComplex = new F2ChainComplex<Integer>();	
	int length = dimension() + 1;
	F2 zero = new F2(0);
	F2 one = new F2(1);
	List<F2Matrix<Integer,Integer>> boundaryOperator 
	    = new ArrayList<F2Matrix<Integer,Integer>>();
	Set<Integer> emptySet = new TreeSet<Integer>();
	if (dimension() >= 0) {
	    Set<Integer> vertices = listOfSimplexSets().get(0);
	    F2Matrix matrix0 = new F2Matrix(emptySet,vertices);
	    boundaryOperator.add(matrix0);
	    for (int i = 1; i < length; i++) {
		F2Matrix<Integer,Integer> matrix 
		    = new F2Matrix<Integer,Integer>();
		TreeMap<ComparablePair<Integer,Integer>,F2> map 
		    = new TreeMap<ComparablePair<Integer,Integer>,F2>();
		Iterator<Integer> rowIter 
		    = listOfSimplexSets().get(i-1).iterator();
		while (rowIter.hasNext()) {
		    Integer rowIndex = rowIter.next();
		    Iterator<Integer> colIter 
			= listOfSimplexSets().get(i).iterator();
		    while (colIter.hasNext()) {
			Integer colIndex = colIter.next();
			Set<K> rowSimplex = inverseMap().get(rowIndex);
			Set<K> colSimplex = inverseMap().get(colIndex);
			ComparablePair<Integer,Integer> pair 
			    = new ComparablePair(rowIndex,colIndex);
			if(isSubSet(rowSimplex,colSimplex)) {
			    map.put(pair,one);
			} else {
			    map.put(pair,zero);
			}
		    }
		}
		matrix = new F2Matrix(map);
		boundaryOperator.add(matrix);
	    }
	}
	chainComplex = new F2ChainComplex(boundaryOperator);
	return chainComplex;
    }
}
