package homology.f2homology;

import java.util.TreeMap;
import java.util.Map;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Set;
import java.util.ArrayList;
import java.util.List;
import topology.*;

/**
 * This class represents a chain complex over the finite field with two 
 * elements.
 * @param <K> index type
 *
 * @author Patrick Erik Bradley
 * @version 1 --- May 26, 2014
 */
@SuppressWarnings("unchecked")
public class F2ChainComplex<K extends Comparable<K>> {
    private int dimension = -1;
    private List<F2Matrix<K,K>> boundaryOperator 
	= new ArrayList<F2Matrix<K,K>>();
    private F2 ZERO = new F2(0);
    
    /**
     * Constructs an empty chain complex.
     */
    public F2ChainComplex() {
    }

    /**
     * Constructs a chain complex from an list of F2Matrices.
     * The first F2Matrix must be one with empty row set. 
     * Throws an exception if the boundary condition for chain complexes 
     * is not satisfied, or the first F2Matrix has non-empty row set.
     *
     * @param <K> index type
     * @param boundaryOperator
     * @throws InvalidOperationException
     */
    public F2ChainComplex(List<F2Matrix<K,K>> boundaryOperator) 
	throws InvalidOperationException {
	
	int N = boundaryOperator.size();
	if (N > 0) {
	    if (boundaryOperator.get(0).rows().size() != 0) {
		throw new InvalidOperationException("Zeroth boundary operator matrix has non-empty row set!");
	    } else {
		//	this.boundaryOperator.add(boundaryOperator.get(0));
	    }
	    for (int i = 0; i < N - 1; i++) {
		
		F2Matrix<K,K> b2 
		    = boundaryOperator.get(i)
		    .product(boundaryOperator.get(i+1));
		if (!boundaryOperator.get(i).rows().isEmpty()) {
		    F2Matrix<K,K> zero 
			= new F2Matrix(boundaryOperator.get(i).rows(),
				   boundaryOperator.get(i+1).columns(),
				       ZERO);
		    
		    if (!b2.equals(zero)) {
			throw new InvalidOperationException("Boundary operator " 
							    + (i)
							    + " times boundary operator " + (i + 1) + " is not zero!");
		    } else {
			this.boundaryOperator.add(boundaryOperator.get(i));
		    }
		} else {
		    this.boundaryOperator.add(boundaryOperator.get(i)); 
		}
	    }
	    this.boundaryOperator.add(boundaryOperator.get(N-1));
	}
	dimension = N - 1;
    }
    
    /**
     * Returns the boundary operator of this chain complex.
     * @param <K> index type
     * @return boundaryOperator
     */
    public List<F2Matrix<K,K>> getBoundaryOperator() {
	return boundaryOperator;
    }

    /**
     * Returns the length of this chain complex.
     * @return length
     */
    public int getDimension() {
	return dimension;
    }

    /**
     * Returns the Betti numbers of the chain complex.
     * @return Betti numbers
     */
    public int[] BettiNumbers() {
	
	int[] betti = new int[dimension + 1];
	
	if (dimension <= 0) {
	    if (dimension == 0) {
	 	betti[0] = boundaryOperator.get(0).columns().size();
	    }
	} else {
	    for (int i = 0; i < dimension; i++) {
		betti[i] = boundaryOperator.get(i).kernelBasis().size()
		    - boundaryOperator.get(i+1).imageBasis().size();
	    }
	    betti[dimension] 
		= boundaryOperator.get(dimension).kernelBasis().size();
	}
	return betti;
    }

    /**
     * Returns the homological dimension of the chain complex.
     * @return homological dimension
     */
    public int homologicalDimension() {
	int dim = -1;
	for (int i = 0; i < dimension + 1; i++) {
	    if (BettiNumbers()[i] != 0) {
		dim = i;
	    }
	}
	return dim;
    }
}
