package homology.f2homology;

import java.lang.Math;
import java.util.TreeMap;
import java.util.Map;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.ListIterator;
import topology.*;
import homology.*;

/**
 * This class represents a matrix over the finite field F2, 
 * viewed as a linear map from F2Vectors
 * parametrised by index type L to F2Vectors parametrised by index type K.
 * It is constructed from a partial matrix by filling in the missing zeros.
 * @param <K,L> the index types
 * 
 * @author Patrick Erik Bradley
 * @version 1.0 --- May 27, 2014
 */
@SuppressWarnings("unchecked")
public class F2Matrix<K extends Comparable<K>,L extends Comparable<L>> 
    extends TreeMap<ComparablePair<K,L>,F2> {

    private F2[][] entries = null;
    private Set<K> rowSet = new TreeSet<K>();
    private Set<L> columnSet = new TreeSet<L>();
    private F2 ZERO = new F2(0);
    private F2 ONE = new F2(1);
    private List<ComparablePair<Integer,Integer>> pivots = new ArrayList();

    /**
     * Creates a new empty F2Matrix.
     */
    public F2Matrix() {
    }

    /**
     * Constructs an empty F2Matrix with specified row set and column set.
     * @param <K> row index type
     * @param <L> column index type
     * @param rowSet
     * @param columnSet
     */
    public F2Matrix(Set<K> rowSet, Set<L> columnSet) {
	this.rowSet = rowSet;
	this.columnSet = columnSet;
	
	entries = new F2[rowSet.size()][columnSet.size()];
    }

    /**
     * Constructs a constant F2Matrix.
     * @param <K> row index type
     * @param <L> column index type
     * @param rowSet
     * @param columnSet
     * @param value
     */
    public F2Matrix(Set<K> rowSet, Set<L> columnSet, F2 value) {
	for (K r : rowSet) {
	    for (L c : columnSet) {
		ComparablePair<K,L> pair = new ComparablePair(r,c);
		put(pair,value);
	    }
	}
	this.rowSet = rowSet;
	this.columnSet = columnSet;
    }

    /**
     * Constructs an F2Matrix from a Map (a partial matrix).
     * @param <K> row index type
     * @param <L> column index type
     * @param map
     */
    public F2Matrix(Map<ComparablePair<K,L>,F2> map) {
	
	// put in the map
	Iterator<ComparablePair<K,L>> iter = map.keySet().iterator();
	while (iter.hasNext()) {
	    ComparablePair<K,L> pair = iter.next();
	    rowSet.add(pair.getKey());
	    columnSet.add(pair.getValue());
	    if (map.get(pair)!=null) {
		put(pair,map.get(pair));
	    } else {
		put(pair,ZERO);
	    }
	}
	
	// fill in with ZEROs
	for (K r : rowSet) {
	    for (L c : columnSet) {
		ComparablePair<K,L> pair = new ComparablePair(r,c);
		if (!keySet().contains(pair)) {
		    put(pair,ZERO);
		}
	    }
	}
    }

    /**
     * Compares the specified F2Matrix with this F2Matrix for 
     * equality in their row sets, column sets and individual entries.
     * @param matrix
     * @return true iff this F2Matrix coincides with matrix
     */
    public boolean equals(F2Matrix<K,L> matrix) {
	boolean equals = false;
	if (rows().equals(matrix.rows()) && columns().equals(matrix.columns())) {
	    equals = true;
	}
	if (equals) {
	    for (K r : rows()) {
	    	for (L c : columns()) {
	    	    ComparablePair<K,L> pair = new ComparablePair(r,c);
	    	    if (get(pair) == null && matrix.get(pair) != null) {
			equals = false;
		    }
		    if (get(pair) != null && matrix.get(pair) == null) {
			equals = false;
		    }
		    if (get(pair) == null && matrix.get(pair) == null) {
			equals = true;
		    }
		    if (get(pair) != null && matrix.get(pair) != null) {
			equals = equals && get(pair).equals(matrix.get(pair));
		    }
	   	}
	    }
	}
	return equals;
    }

    /**
     * Compares the specified F2Matrix with this F2Matrix for equality in 
     * their arrays.
     * @param matrix
     * @return true iff the array of matrix coincides 
     * with the array of this F2Matrix
     */
    public boolean equalArray(F2Matrix<K,L> matrix) {
	boolean equals = false;
	int R = rows().size();
	int C = columns().size();
	if (matrix.rows().size() == R && matrix.columns().size() == C) {
	    equals = true;
	}
	if (equals) {
	    for (int i = 0; i < R; i++) {
	    	for (int j = 0; j < C; j++) {
	    	    equals = equals && toArray()[i][j]
	    		.equals(matrix.toArray()[i][j]);
	    	}
	    }
	}
	return equals;
    }

    /**
     * Returns the row set of this F2Matrix.
     * @param <K> row index type
     * @return row set 
     */
    public Set<K> rows() {
	return rowSet;
    }

    /**
     * Returns the column set of this F2Matrix.
     * @param <L> column index type
     * @return column set
     */
    public Set<L> columns() {
	return columnSet;
    }

    /**
     * Returns a list of row indices of this F2Matrix.
     * @param <K> row index type
     * @return rowList
     */
    public List<K> rowList() {
	List<K> rowList = new ArrayList();
	for (K r : rowSet) {
	    rowList.add(r);
	}
	return rowList;

    }

    /**
     * Returns a list of column indices of this F2Matrix.
     * @param <L> column index type
     * @return columnList
     */
    public List<L> columnList() {
	List<L> columnList = new ArrayList();
	for (L c : columnSet) {
	    columnList.add(c);
	}
	return columnList;
    }

    /**
     * Returns the array associated with this F2Matrix.
     * @return array
     */
    public F2[][] toArray() {

	List<K> rowList = rowList(); 
	List<L> columnList = columnList(); 
	int rowSize = rowList.size(); 
	int columnSize = columnList.size();
	entries = new F2[rowSize][columnSize];
	for(int i = 0; i < rowSize; i++) {
	    K r = rowList.get(i);
	    for(int j = 0; j < columnSize; j++) {
		L c = columnList.get(j); 
		ComparablePair<K,L> pair = new ComparablePair(r,c);
		if (get(pair) != null) {
		    entries[i][j] = get(pair); 
		} else {
		    entries[i][j] = ZERO;
		}
	    }
	}
    	return entries;
    }

    /**
     * Returns an upper triangular array by partial Gaussian elimination
     * of this F2Matrix.
     * @param 
     */
    public F2[][] upperTriangArray() {
	int R = rowSet.size(); 
	int C = columnSet.size();
	int N = Math.min(R,C);
	F2[][] upperTriangArray = new F2[R][C]; 
	upperTriangArray = toArray(); 
	if (R > 0 && C > 0) {
	    int pRow = 0; 
	
	    for (int p = 0; p < C; p++) {
	    
		// find pivot row 
		int max = pRow;
		for (int i = pRow + 1; i < R; i++) {
		    if (upperTriangArray[i][p].equals(ONE)) {
			max = i;
			break;
		    } 
		}
		
		// if pivot found, swap
		if (upperTriangArray[max][p].equals(ONE)) {
		    
		    F2[] temp = upperTriangArray[pRow]; 
		    upperTriangArray[pRow] = upperTriangArray[max]; 
		    upperTriangArray[max] = temp;
		    
		    // eliminate below pivot
		    for (int i = pRow + 1; i < R; i++) {
			F2 alpha = upperTriangArray[i][p];
			for (int j = p; j < C; j++) {
			    upperTriangArray[i][j] 
				= upperTriangArray[i][j]
				.add(alpha.multiply(upperTriangArray[pRow][j]));
			}
		    }
		    
		    if (pivots.size() < N) {
			pivots.add(new ComparablePair(pRow,p));
		    }
		    if (pRow < R - 1) {
			pRow = pRow + 1;
		    }
		}
	    }
	}
	return upperTriangArray;
    }

    /**
     * Returns an upper triangular matrix by partial 
     * Gaussian elimination of this F2Matrix.
     * @param <K> row index type
     * @param <L> column index type
     */
    public F2Matrix<K,L> upperTriangular() {
	F2Matrix<K,L> upperTriangular = new F2Matrix();
	TreeMap<ComparablePair<K,L>,F2> UTMap = new TreeMap();
	F2[][] U = upperTriangArray();
	int R = rows().size();
	int C = columns().size();
	for (int i = 0; i < R; i++) {
	    K r = rowList().get(i);
	    for (int j = 0; j < C; j++) {
		L c = columnList().get(j);
		UTMap.put(new ComparablePair(r,c), U[i][j]);
	    }
	}
	upperTriangular = new F2Matrix(UTMap);
	return upperTriangular;
    }

    /**
     * Returns an array which is a row echelon form of this F2Matrix.
     */
    public F2[][] rowEchelonArray() {
	int R = rows().size();
	int C = columns().size();
	F2[][] rowEchelonArray = new F2[R][C];
	rowEchelonArray = upperTriangArray();
	int P = pivots.size();
	for (int p = 0; p < P; p++) {
	    Pair<Integer,Integer> pivot = pivots.get(p);
	    int i = pivot.getKey();
	    int j = pivot.getValue();

	    // eliminate above pivot
	    for (int k = i - 1; k >= 0; k--) {
		F2 alpha = rowEchelonArray[k][j];
		for (int l = j; l >= 0; l--) {
		    rowEchelonArray[k][l] =
			rowEchelonArray[k][l]
			.add(alpha.multiply(rowEchelonArray[i][l]));
		}
	    }
	    
	}
	return rowEchelonArray;
    }

    /**
     * Returns a row echelon form of this F2Matrix.
     * @param <K> row index type
     * @param <L> column index type
     */
    public F2Matrix<K,L> rowEchelon() {
	F2Matrix<K,L> rowEchelon = new F2Matrix();
	TreeMap<ComparablePair<K,L>,F2> rEMap = new TreeMap();
	F2[][] E = rowEchelonArray();
	int R = rows().size();
	int C = columns().size();
	for (int i = 0; i < R; i++) {
	    K r = rowList().get(i);
	    for (int j = 0; j < C; j++) {
		L c = columnList().get(j);
		rEMap.put(new ComparablePair(r,c), E[i][j]);
	    }
	}
	rowEchelon = new F2Matrix(rEMap);
	return rowEchelon;
    }

    /**
     * Returns a standard array over F2 having a one at the given place,
     * and zero otherwise. Its length is the number of columns.
     * @param place  
     * @return standardRow
     */
    private F2[] standardRow(int place) {
	int C = columns().size();
	F2[] standardRow = new F2[C];
	if (place < 0 || place >= C) {
	    throw new InvalidOperationException("place = " + place 
						+ "not within row of this matrix!");
	} else {
	    for (int i = 0; i < C; i++) {
		if (i == place) {
		    standardRow[i] = ONE;
		    }
		else {
		    standardRow[i] = ZERO;
		}
	    }
	}
	return standardRow;
    }

    /**
     * Returns the list of pivot column indices.
     * @return pivotColumns
     */
    private List<Integer> pivotColumns() {
	List<Integer> pivotColumns = new ArrayList();
	for (ComparablePair<Integer,Integer> pivot : pivots) {
	    int pc = pivot.getValue();
	    pivotColumns.add(pc);
	}
	return pivotColumns;
    }

    /**
     * Returns the normal form for reading off a basis for the kernel of this
     * F2Matrix. The result is a list of column arrays.
     * @return readOffNF
     */
    public List<F2[]> readOffNF() {
	int N = rows().size();
	List<F2[]> readOffNF = new ArrayList(Arrays.asList(rowEchelonArray()));

	// find last pivot
	int maxRow = 0;
	if (pivots.size() > 0) {
	    //ComparablePair<Integer,Integer> maxPivot 
	    //	= pivots.get(pivots.size() - 1);
	    for (Pair<Integer,Integer> pair : pivots) {
		Integer key = pair.getKey();
		if (key > maxRow) {
		    maxRow = key; 
		    }
	    }
	    //  maxRow = maxPivot.getKey();
	}
		
	// remove the zero rows
	if (maxRow < N - 1) {
	    for (int i = N-1; i > maxRow; i--) {
		readOffNF.remove(i);
	    }
	}
	
	// insert standard rows
	int M = columns().size();
	for (int p = 0; p < M; p++) {
	    if (!pivotColumns().contains(p)) {
		readOffNF.add(p,standardRow(p));
	    } 
	}
		
	// transform to list of column arrays
	F2[][] readOffTranspose = new F2[M][M];
	for (int i = 0; i < M; i++) {
	    for (int j = 0; j < M; j++) {
		readOffTranspose[i][j] 
		    = new F2(readOffNF.get(j)[i].getNumber());
	    }
	}
	readOffNF = new ArrayList(Arrays.asList(readOffTranspose)); 
	return readOffNF;
    }
    
    /** 
     * Returns a basis for the kernel of this F2Matrix
     * as a list of arrays over F2.
     * @return kernelBasisListOfArray 
     */
    public List<F2[]> kernelBasisListOfArray() {
	List<F2[]> kernelBasisListOfArray = new ArrayList();
	List<F2[]> readOffNF = readOffNF();
	int N = readOffNF.size();
	for (int p = 0; p < N; p++) {
	    if (!pivotColumns().contains(p)) {
		kernelBasisListOfArray.add(readOffNF.get(p));
	    }
	}
	return kernelBasisListOfArray;
    }
    
    /**
     * Returns a basis for the kernel of this F2Matrix 
     * as a list of F2Vectors.
     * @param <L> column index type
     * @return kernelBasis
     */
    public List<F2Vector<L>> kernelBasis() {
	List<F2Vector<L>> kernelBasis =  new ArrayList();
	List<F2[]> kernelBasisListOfArray = kernelBasisListOfArray();
	int N = kernelBasisListOfArray.size();
	int C = columns().size();
	for (int i = 0; i < N; i++) {
	    TreeMap<L,F2> columnMap = new TreeMap();
	    for (int j = 0; j < C; j++) {
		L c = columnList().get(j);
		columnMap.put(c, kernelBasisListOfArray.get(i)[j]);
	    }
	    F2Vector<L> column = new F2Vector(columnMap);
	    kernelBasis.add(column);
	} 
	return kernelBasis;
    }
    
    /**
     * Returns a basis of the image of this F2Matrix as a list of arrays.
     * @return imageBasisListOfArray
     */ 
    public List<F2[]> imageBasisListOfArray() {
	List<F2[]> imageBasisListOfArray = new ArrayList();
	F2[][] upperTriangArray = upperTriangArray();
	int C = columns().size();
	int R = rows().size();
	for (int p = 0; p < C; p++) {
	    if (pivotColumns().contains(p)) {
	    	F2[] column = new F2[R];
	    	for (int i = 0; i < R; i++) {
	    	        column[i] = toArray()[i][p];
	    	}
	    	imageBasisListOfArray.add(column);
	    }
	}
	return imageBasisListOfArray;
    }
    
    /**
     * Returns a basis of F2Vectors for the image of this F2Matrix. 
     * @param <K> row index type
     * @return imageBasis
     */
    public List<F2Vector<K>> imageBasis() {
	List<F2Vector<K>> imageBasis = new ArrayList();
	List<F2[]> imageBasisListOfArray = imageBasisListOfArray();
	int N = imageBasisListOfArray.size();
	int R = rows().size();
	for (int i = 0; i < N; i++) {
	    TreeMap<K,F2> columnMap = new TreeMap();
	    for (int j = 0; j < R; j++) {
		K r = rowList().get(j); 
		columnMap.put(r,imageBasisListOfArray.get(i)[j]);
	    }
	    F2Vector<K> column = new F2Vector(columnMap);
	    imageBasis.add(column);
	}
	return imageBasis;
    }
    
    /**
     * Returns the rank of this F2Matrix.
     * @return rank
     */
    public int rank() {
	int rank = imageBasisListOfArray().size();
	return rank;
    }

    /**
     * Returns the product AB of this matrix A with another matrix B.
     * @param <K> row index type
     * @param <L> row index type
     * @param <M> column index type
     * @param m
     * @return product with m (the tree matrix B)
     * @throws InvalidOperationException if the column set of A does not coincide with the row set of B
     */
    public <M extends Comparable<M>> F2Matrix<K,M> product(F2Matrix<L,M> m) {
	
	F2Matrix<K,M> product = new F2Matrix();
	TreeMap<ComparablePair<K,M>,F2> productMap = new TreeMap();
	if (!columns().equals(m.rows())) {
	    throw new InvalidOperationException("Tried to multiply a matrix with " + columns().size() + " columns and a matrix with " 
						+ m.rows().size() + " rows");

	} else {
	Set<K>    resultRows = rows();
	Set<M>    resultColumns = m.columns();
	Iterator<K> rowIter = resultRows.iterator();
	while (rowIter.hasNext()) {
	    K a = rowIter.next();
	    Iterator<M> colIter = resultColumns.iterator();
	    while (colIter.hasNext()) {
		M b = colIter.next();
		ComparablePair<K,M> pair = new ComparablePair(a,b);
		F2 value = ZERO;
		for (L c : columns()) {
		    ComparablePair<K,L> ac = new ComparablePair(a,c);
		    ComparablePair<L,M> cb = new ComparablePair(c,b);
		    value = value.add(get(ac).multiply(m.get(cb)));
		}
		productMap.put(pair,value);
	    }
	    product = new F2Matrix(productMap);
	}
	}
	return product;
    }
}
