package homology.f2homology;

import java.util.TreeMap;
import java.util.Map;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;

import topology.*;

/**
 * This class represents a module over the finite field with two elements, 
 * i.e. a vector space.
 * @param <K> the index type
 *
 * @author Patrick Erik Bradley
 * @version 1.0 --- April 10, 2014
 */
@SuppressWarnings("unchecked")
public class F2Module<K extends Comparable<K>> {
    
    private int rank = -1;
    private List<F2Vector<K>> generatingSystem = new ArrayList();
    private F2Matrix<K,Integer> matrix = new F2Matrix();
    private Set<K> keySet = new TreeSet();
    private F2 ZERO = new F2(0);

    /**
     * Constructs a new empty F2Module.
     */
    public F2Module() {
    }

    /**
     * Constructs a module (in fact a vector space) with given 
     * generating System of F2Vectors;
     * @param <K> index type
     * @param generatingSystem
     */
    public F2Module(List<F2Vector<K>> generatingSystem) {
	
	int L = generatingSystem.size();
	int N = generatingSystem.get(0).length();
	for (int i = 0; i < L; i++) {
	    if (generatingSystem.get(i).length() != N) {
		throw new InvalidOperationException("The F2Vectors have different lengths!");
	    }
	}
	if (L > 0 && N > 0) {
	    for (int i = 0; i < L; i++) {
		keySet.addAll(generatingSystem.get(i).keySet());
		F2Vector<K> vector = new F2Vector();
		for (K k : keySet) {
		    if (generatingSystem.get(i).get(k) == null) {
			vector.put(k,ZERO);
		    } else {
			vector.put(k,generatingSystem.get(i).get(k));
		    }
		}
		this.generatingSystem.add(vector);
	    }
	}
	else {
	    this.generatingSystem = generatingSystem;
	}
    }

    /**
     * Returns the generating system of this F2Module.
     * @param <K> index type
     * @return generatingSystem
     */
    public List<F2Vector<K>> getGeneratingSystem() {
	return generatingSystem;
    }

    /**
     * Returns the F2Matrix associated with the generating system
     * of this F2Module.
     * @param <K> index type
     * @return matrix
     */
    public F2Matrix<K,Integer> getMatrix() {
	int N = generatingSystem.size();
	TreeMap<ComparablePair<K,Integer>,F2> map = new TreeMap();
	for (K r : keySet) {
	    for (int c = 0; c < N; c++) {
		ComparablePair<K,Integer> pair = new ComparablePair(r,c);
		F2 value = generatingSystem.get(c).get(r);
		map.put(pair,value);
	    }
	}
	matrix = new F2Matrix(map);
	return matrix;
    }

    /**
     * Returns a basis of the module.
     * @param <K> index type
     * @return basis
     */
    public List<F2Vector<K>> basis() {
	List<F2Vector<K>> basis = getMatrix().imageBasis();
	return basis;
    } 

    /**
     * Returns the rank of the module.
     * @return rank 
     */
    public int getRank() {
	rank = basis().size();
	return rank;
    }
}
