package homology.f2homology;

import homology.*;
import homology.f2homology.*;
import topology.*;
import java.util.*;

/**
 * This class computes the quotient of the order complex of a
 * finite space with the order complex of a second space whose
 * point set coincides with the point set of the first space,
 * and whose Hasse diagram is contained in the Hasse diagram
 * of the first space. This is called a special submersion.
 * @param <E> space type
 * @author Patrick Erik Bradley
 * @version 1.0 --- May 30, 2014
 */
public class F2OrderComplexQuotient<E extends Comparable<E>> {
    
    private F2ChainComplex<Integer> F2ChainComplex = new F2ChainComplex();
    private FiniteSpace<E> space1 = new FiniteSpace<E>();
    private FiniteSpace<E> space2 = new FiniteSpace<E>();
    private OrderComplex<E> orderComplex1 = new OrderComplex();
    private OrderComplex<E> orderComplex2 = new OrderComplex();
    
    /**
     * Creates a new instance of F2OrderComplexQuotient
     */
    public F2OrderComplexQuotient() {
    }

    /**
     * Creates the quotient complex of space1 and space2 if
     * space1 is specially submersed in space2
     * @param <E> space type
     * @param space1
     * @param space2
     */
    public F2OrderComplexQuotient(FiniteSpace<E> space1, 
				FiniteSpace<E> space2) {
	this.space1 = new FiniteSpace(space1);
	this.space2 = new FiniteSpace(space2);
	
	if (isSpecialSubmersion()) {
	    orderComplex1 = new OrderComplex(space1);
	    orderComplex2 = new OrderComplex(space2);
	      
	    // create a list of simplex sets by removing simplices
	    List<TreeSet<Integer>> listOfSimplexSets = listOfSimplexSets();

	    // create chain complex over finite field F2
	    int length = orderComplex2.dimension() + 1;
	    F2 zero = new F2(0);
	    F2 one = new F2(1);
	    List<F2Matrix<Integer,Integer>> boundaryOperator 
		= new ArrayList<F2Matrix<Integer,Integer>>();
	    Set<Integer> emptySet = new TreeSet<Integer>();
	    if (length > 0) {
		Set<Integer> vertices = listOfSimplexSets.get(0);
		F2Matrix matrix0 = new F2Matrix(emptySet,vertices);
		
		if (length > 1) {
		    Set<Integer> lines = listOfSimplexSets.get(1);
		    F2Matrix matrix1 = new F2Matrix(vertices,lines,zero);
		    
		    for (int i = 2; i < length; i++) {
			F2Matrix<Integer,Integer> matrix 
			    = new F2Matrix<Integer,Integer>();
			TreeMap<ComparablePair<Integer,Integer>,F2> map 
			    = new TreeMap<ComparablePair<Integer,Integer>,F2>();
			Iterator<Integer> rowIter 
			    = listOfSimplexSets.get(i-1).iterator();
			while (rowIter.hasNext()) {
			    Integer rowIndex = rowIter.next();
			    Iterator<Integer> colIter 
			    = listOfSimplexSets.get(i).iterator();
			    while (colIter.hasNext()) {
				Integer colIndex = colIter.next();
				Set<E> rowSimplex 
				    = orderComplex2.inverseMap().get(rowIndex);
				Set<E> colSimplex 
				= orderComplex2.inverseMap().get(colIndex);
				ComparablePair<Integer,Integer> pair 
				    = new ComparablePair(rowIndex,colIndex);
				if(colSimplex.containsAll(rowSimplex)) {
				    map.put(pair,one);
				} else {
				    map.put(pair,zero);
				}
			    }
			}
			matrix = new F2Matrix(map);
			boundaryOperator.add(matrix0);
			boundaryOperator.add(matrix1);
			boundaryOperator.add(matrix);
		    }
		} else {
		    boundaryOperator.add(matrix0);
		}
		
	    }
	    F2ChainComplex = new F2ChainComplex(boundaryOperator);
	}
    }
    
    /**
     * Returns the list of simplex sets.
     */
    public List<TreeSet<Integer>> listOfSimplexSets() {

	Set<Set<E>> simplices1 = orderComplex1.simplices();
	Set<Set<E>> simplices2 = orderComplex2.simplices();
	Set<Set<E>> simplicesToRemove 
	    = new TreeSet<Set<E>>(new TreeSetComparator());
	for (Set<E> simplex : simplices2) {
	    if (simplices1.contains(simplex)) {
		simplicesToRemove.add(simplex);
	    }
	}
	List<TreeSet<Integer>> listOfSimplexSets = new ArrayList();
	listOfSimplexSets.addAll(orderComplex2.listOfSimplexSets());
	for (Set<E> simplex : simplicesToRemove) {
	    int s = simplex.size();
	    int index = orderComplex2.indexMap().get(simplex);
	    listOfSimplexSets.get(s-1).remove(index);
	}
	return listOfSimplexSets;
    }

    /**
     * Returns true iff there is a special submersion from space1 into space2.
     */
    public boolean isSpecialSubmersion() {
	boolean isSpecialSubmersion = false;
	if (space2.getPointSet().equals(space1.getPointSet())
	    && space2.HasseDiagram().getPairs()
	    .containsAll(space1.HasseDiagram().getPairs())) {
	    isSpecialSubmersion = true;
	}
	return isSpecialSubmersion;
    }

    /**
     * Returns the quotient complex over the finite field F2.
     */
    public F2ChainComplex<Integer> getF2ChainComplex() {
	return F2ChainComplex;
    }
    
}
