package topology;

import java.util.*;

/**
 * The class BinaryRelation defines a set of pairs as a binary relation
 * on the domain defined by the pairs themselves.
 * @author Patrick Erik Bradley
 * @version 1.0
 */
public class BinaryRelation<E extends Comparable<E>> {

    /**
     * the set of pairs making up the relation
     * @param <E> relation type
     */
    private Set<Pair<E,E>> pairs;
    
    /**
     * Creates an empty binary relation. 
     */
    public BinaryRelation() {
	this.pairs = new TreeSet();
    }
    
    /** 
     * Creates a new BinaryRelation given by a set of pairs.
     * @param <E> relation type
     * @param a set of pairs
     */
    public BinaryRelation(Set<Pair<E,E>> somePairs) {
	this.pairs = somePairs;
    }

    /**
     * Creates a new BinaryRelation given by a BinaryRelation
     * @param <E> relation type
     * @param a relation
     */
    public BinaryRelation(BinaryRelation<E> relation) {
	this.pairs = relation.pairs;
    }

    /**
     * Returns the pairs of this BinaryRelation.
     * @param <E> relation type
     * @return set of pairs
     */
    public Set<Pair<E,E>> getPairs() {
	return pairs;
    }
    
    /**
     * Adds a pair to the pairs of this BinaryRelation.
     * @param <E> relation type
     * @param pair
     * @return true iff the pair does not previously belong to this relation
     */
    public boolean add(Pair<E,E> pair) {
	return pairs.add(pair);
    }
    
    /**
     * Adds a set of pairs to the pairs of this BinaryRelation.
     * @param <E> relation type
     * @param pairSet
     * @return true iff all pairs in the set do not previously belong to this
     * relation 
     */
    public boolean addAll(Set<Pair<E,E>> pairSet) {
	return pairs.addAll(pairSet);
    }

    /**
     * Removes a pair from the pairs of this BinaryRelation.
     * @param <E> relation type
     * @param pair
     * @return true iff the pair previously belongs to this relation
     */
    public boolean remove(Pair<E,E> pair) {
	return pairs.remove(pair);
    }

    /**
     * Removes a set of pairs from the pairs of this BinaryRelation.
     * @param <E> relation type
     * @param pairSet
     * @return true iff all pairs in the set previously belong to this
     * relation
     */ 
    public boolean removeAll(Set<Pair<E,E>> pairSet) {
	return pairs.removeAll(pairSet);
    }
    
    /**
     * Returns all elements in this BinaryRelation to a given element.
     * @param <E> relation type
     * @param e an element
     * @return the set of all elements in relation with given element
     */
    public Set<E> inLink(E e) {
	Set<E> link = new TreeSet();
	Iterator iter = pairs.iterator();
	while (iter.hasNext()){
	    Pair<E,E> curr = (Pair<E,E>)iter.next(); 
	    if (e.equals(curr.getValue())) {
		link.add(curr.getKey());
	    }
	}
	return link;
    }
    
    /**
     * Returns all elements to which a given element is in this 
     * BinaryRelation with.
     * @param <E> relation type
     * @param e an element
     * @return the set of all elements the given element is in relation with
     */
    public Set<E> outLink(E e) {
	Set<E> link = new TreeSet();
	Iterator iter = pairs.iterator();
	while (iter.hasNext()){
	    Pair<E,E> curr = (Pair<E,E>)iter.next();
	    if (e.equals(curr.getKey())) {
		link.add(curr.getValue());
	    }
	}
	return link;
    }

    /**
     * Returns all pairs (e,e) where e is in BinaryRelation with some element
     * or some element is in BinaryRelation with e.
     * @param <E> relation type
     * @return set of pairs whose key and value are identical
     */
    public Set<Pair<E,E>> diagonal() {
	Set<Pair<E,E>> diag = new TreeSet();
	Iterator iter = pairs.iterator();
	while (iter.hasNext()) {
	    Pair<E,E> it = (Pair<E,E>)iter.next();
	    if (it.getKey() == it.getValue()) {
		diag.add(it);
		} else { 
		Pair<E,E> p = new ComparablePair(it.getKey(),it.getKey());
		diag.add(p);
		Pair<E,E> q = new ComparablePair(it.getValue(),it.getValue());
		diag.add(q);
	    }
	}
	return diag;
    }

    /**
     * Returns true if BinaryRelation is anti-symmetric. 
     * Otherwise returns false.
     * @return true iff this relation is anti-symmetric
     */
    public boolean isAntiSymmetric() {
	boolean result = true;
	Iterator iter = pairs.iterator();
	while (iter.hasNext()) {
	    Pair<E,E> currPair = (Pair<E,E>)iter.next();
	    E a = currPair.getKey();
	    E b = currPair.getValue();
	    if (!a.equals(b)) {
		Pair<E,E> rrucPair = new ComparablePair(b, a);
		if (pairs.contains(rrucPair)) {
		    result = false;
		    break;
		} 
	    }
	}
	return result;
    }

    /**
     * Checks equality of this BinaryRelation with another BinaryRelation.
     * @param F relation type
     * @param relation
     * @return true iff this relation coincides with given relation
     */
    public <F extends Comparable<F>> boolean equals(BinaryRelation<F> relation) {
	return this.pairs.equals(relation.pairs);
    }
}
