package topology;

import java.util.*;

/**
 * Implements the mathematical object named Finite Topological Space.
 * @author Patrick Erik Bradley
 * @version 1.0
 */
public class FiniteSpace<E extends Comparable<E>> implements TopologicalSpace<E>
{
    private Set<E> pointSet = new TreeSet();
    private BinaryRelation<E> relation = new BinaryRelation<E>();
   
    /**
     * Creates an empty space.
     */
    public FiniteSpace() {
	pointSet = new TreeSet();
	relation = new BinaryRelation();
    }

    /**
     * Creates a space with given point set and binary relation.
     * Pairs with key or value not in the point set are ignored,
     * and are not part of the topology defining relation.
     * @param <E> space type
     * @param points a point set
     * @param rel a relation
     */
    public FiniteSpace(Set<E> points, BinaryRelation<E> rel) {
	pointSet = points;
	BinaryRelation<E> relatie = new BinaryRelation<E>(rel);
	Iterator<Pair<E,E>> iter = relatie.getPairs().iterator();
	while (iter.hasNext()) {
	    Pair<E,E> pair = (Pair<E,E>)iter.next();
	    if (!points.contains(pair.getKey()) 
		|| !points.contains(pair.getValue())) {
		iter.remove();
	    }
	}
	relation = new BinaryRelation(relatie);
    } 
    
    /**
     * Creates a new finite topological space from a given finite space. 
     * @param <E> space type
     * @param space
     */
    public FiniteSpace(FiniteSpace<E> space) {
	this.addAllPoint(space.getPointSet());
	this.addAllPair(space.getRelation().getPairs());
    }

    /**
     * Returns the point set underlying the topological space.
     * @return point set
     */
    public Set<E> getPointSet() {
	return pointSet; 
    }
    
    /**
     * Returns the topology-defining relation.
     * @return relation
     */    
    public BinaryRelation<E> getRelation() {
	return relation; 
    }
    
    /** 
     * Checks equality of this finite topological space with a given
     * finite topological space. 
     * @param <F> space type
     * @param space
     * @return true iff this space coincides with given space
     */
    public <F extends Comparable<F>> boolean equals(FiniteSpace<F> space) {
	return (this.pointSet.equals(space.pointSet) 
		&& this.relation.equals(space.relation));
    }

    /**
     * Returns the reflexive and transitive closure of the 
     * topology-defining relation.
     * @param <E> relation type
     * @return pre-order
     */
    public BinaryRelation<E> getPreOrder() {
	BinaryRelation<E> preOrder = new BinaryRelation();        
	
	// add the diagonal
	Iterator<E> itPoints = pointSet.iterator();
	while (itPoints.hasNext()) {
	    E it = itPoints.next();
	    Pair<E,E> p = new ComparablePair(it,it);
	    preOrder.add(p);
	}
	
	// add the transitive closure of the relation
	BinaryRelation<E> rel = new BinaryRelation();
	rel.addAll(relation.getPairs());
	BinaryRelation<E> tc = RelationCollection.transitiveClosure(rel);
	preOrder.addAll(tc.getPairs());
	
	return preOrder;
    }

    /**
     * Returns true iff pre-order is antisymmetric. Otherwise, returns false.
     * @return true iff this space is T0
     */
    public boolean isTzero() {
	BinaryRelation<E> preOrder = this.getPreOrder();
	boolean isTnull = preOrder.isAntiSymmetric();
	return isTnull;
    }

    /**
     * Adds a point to the point set of a finite topological space.
     * @param e a point
     * @return true iff point does not previously belong to this space
     */
    public boolean addPoint(E e) {
	return pointSet.add(e);
    }

    /**
     * Adds a set of points to the point set of a finite topological space.
     * @param pts a set of points
     * @return true iff all elements of the set do not previously 
     * belong to this space
     */
    public boolean addAllPoint(Set<E> pts) {
	return pointSet.addAll(pts);
    }

    /**
     * Adds a pair to the topology-defining relation of a 
     * finite topological space, if the pair's key and value 
     * belong to the point set. 
     * Otherwise, a warning is issued.
     * @param e a pair
     * @return true iff pair does not previously belong to this space, but key
     * and value do
     */
    public boolean addPair(Pair<E,E> e) {
	E e1 = e.getKey();
	E e2 = e.getValue();
	if (pointSet.contains(e1) & pointSet.contains(e2)){ 
	    return relation.add(e);
	} else {
	    System.out.println("Warning: Pair (" + e1 + "," + e2 +") not on pointSet.");
	    return false;
	}
    }
    
    /**
     * Adds a set of pairs to the topology-defining relation of a finite
     * topological space, if the pairs' key and value belong to the
     * point set. Issues a warning for those pairs for which that is not
     * the case.
     * @param s a set of pairs
     * @return true iff all elements of the set do not previously belong to this
     * space, but keys and values do 
     */
    public boolean addAllPair(Set<Pair<E,E>> s) {
	boolean r = true;
	for (Pair<E,E> e : s) {
	    r = r && this.addPair(e);
	}
	return r;
    }
    
    /**
     * Removes a pair from the topology-defining relation.
     * @param e a pair
     * @return true iff pair previously belongs to space
     */
    public boolean removePair(Pair<E,E> e) {
	return relation.remove(e);
    }

    /**
     * Removes a set of pairs from the topology-defining relation.
     * @param s a set of pairs
     * @return true iff all elements of the set previously belong to this space
     */
    public boolean removeAllPair(Set<Pair<E,E>> s) {
	return relation.removeAll(s);
    }

    /**
     * Removes a point from the point set, and updates the relation.
     * @param e a point
     * @return true iff point previously belongs to this space
     */
    public boolean removePoint(E e){
	if (pointSet.contains(e)) {
	    Set<E> in = relation.inLink(e);
	    Set<E> out = relation.outLink(e);
	    Iterator inIter = in.iterator();
	    while (inIter.hasNext()) {
		E inNext = (E)inIter.next(); 
		Iterator outIter = out.iterator();
		while (outIter.hasNext()){
		    E outNext = (E)outIter.next();
		    Pair<E,E> p = new ComparablePair(inNext,outNext);
		    relation.add(p);
		}
	    }
	    inIter = in.iterator();
	    while (inIter.hasNext()) {
		Pair<E,E> p = new ComparablePair((E)inIter.next(),e);
		relation.remove(p);
	    }
	    Iterator outIter = out.iterator();
	    while (outIter.hasNext()) {
		Pair<E,E> p = new ComparablePair(e,(E)outIter.next());
		relation.remove(p);
	    }
	}
	return pointSet.remove(e);
    }
    
    /**
     * Removes a set of points and updates the relation.
     * @param s a set of points
     * @return true iff all elements of the set previously belong to this space
     */
    public boolean removeAllPoint(Set<E> s) {
	boolean r = true;
	for(E e : s) {
	    r = r && this.removePoint(e);
	}
	return r;
    }

    /**
     * Returns all paths from a source to a destination point.
     * @param source
     * @param destination
     * @return all paths from source to destination
     */
    public Set<List<E>> allPaths(E source, E destination) {
	Set<List<E>> paths = new TreeSet<List<E>>(new ListComparator<E>());
	recursive(source, destination, paths, new ArrayList<E>());
	return paths;
    }

    /**
     * Recursively computes paths
     * @param current current point
     * @param destination
     * @param paths a set of paths
     * @param path
     */
    private void recursive(E current, E destination, 
			   Set<List<E>> paths, ArrayList<E> path) {
	path.add(current);
	
	if (destination.equals(current)) {
	    //paths.add(path);
	    paths.add((List<E>)path.clone());
	    path.remove(current);
	    return;
	}
	
	for (E e : this.getRelation().outLink(current)) {
	    if (!path.contains(e)) {
		recursive(e, destination, paths, path);
	    }
	}
	
	path.remove(current);
    }

    /**
     * Returns the set of fat points of this finite topological space.
     * These are those equivalence classes containing at least two elements,
     * for the equivalence relation  given by x above y and y above x, 
     * where "above" denotes the pre-order induced by the topology defining 
     * relation. 
     * @return set of fat points
     */
    public Set<TreeSet<E>> fatPoints() {
	Set<TreeSet<E>> fatPoints 
	    = new TreeSet<TreeSet<E>>(new TreeSetComparator());
	Iterator<E> iter = getPointSet().iterator();
	while (iter.hasNext()) {
		E point = iter.next();
		TreeSet<E> fatPoint = new TreeSet<E>();
		//fatPoint.add(point);
		Set<E> above = getPreOrder().inLink(point);
		Set<E> below = getPreOrder().outLink(point);
		Iterator<E> aboveIter = above.iterator();
		while (aboveIter.hasNext()) {
		    E a = aboveIter.next(); 
		    if (!a.equals(point)) {
			Pair<E,E> riap = new ComparablePair(point,a);
			if (getPreOrder().getPairs().contains(riap)) {
			    fatPoint.add(point);
			    fatPoint.add(a);
			}
		    }
		}
		Iterator<E> belowIter = below.iterator();
		while (belowIter.hasNext()) {
		    E b = belowIter.next();
		    if (!b.equals(point)) {
			Pair<E,E> riap = new ComparablePair(b,point);
			if (getPreOrder().getPairs().contains(riap)) {
			    fatPoint.add(point);
			    fatPoint.add(b);
			}
		    }
		}
		if (!fatPoint.isEmpty()) {
		    fatPoints.add(fatPoint);
		}
	    }
	return fatPoints;
    }

    /**
     * Returns the Kolmogorov Quotient of this finite topological space
     * by removing points.
     * @param <E> space type
     * @return Kolmogorov quotient
     */
    public FiniteSpace<E> KolmogorovQuotient() {
	
	FiniteSpace<E> result 
	    = new FiniteSpace(getPointSet(), getRelation()); 
	if (!isTzero()) {
	    Set<TreeSet<E>> fatPoints = fatPoints();
	    Iterator<TreeSet<E>> fatIter = fatPoints.iterator();
	    while (fatIter.hasNext()) {
		TreeSet<E> fatPoint = fatIter.next();
		E first = fatPoint.first();
		Iterator<E> iter = fatPoint.iterator();
		while (iter.hasNext()) {
		    E point = iter.next();
		    if (!point.equals(first)) {
			result.removePoint(point);
		    }
		}
	    }
	}
	return result;
    }

    /**
     * Returns the Alexandrov dual of this finite topological space.
     * @param <E> space type
     * @return dual space
     */
    public FiniteSpace<E> dual() {
	FiniteSpace<E> result = new FiniteSpace();
	BinaryRelation<E> noitaler = new BinaryRelation();
	Iterator<Pair<E,E>> iter = getRelation().getPairs().iterator();
	while (iter.hasNext()) {
	    Pair<E,E> pair = iter.next();
	    E a = pair.getKey();
	    E b = pair.getValue();
	    if (a != b) {
		Pair<E,E> riap = new ComparablePair(b,a);
		noitaler.add(riap);
	    }
	}
	result = new FiniteSpace(getPointSet(),noitaler);
	return result;
    }

    /**
     * Returns  the Hasse diagram of the Kolmogorov quotient of this finite
     * topological space.
     * @param <E> space type
     * @return Hasse diagram
     */
    public BinaryRelation<E> HasseDiagram() {
	BinaryRelation<E> result 
	    = new BinaryRelation(KolmogorovQuotient().getRelation());
	result.removeAll(result.diagonal());
	Iterator <E> iter = KolmogorovQuotient().getPointSet().iterator();
	while (iter.hasNext()) {
	  E a = iter.next();
	  Set<E> outSet1 = result.outLink(a);
	  Iterator <E> outer = outSet1.iterator();
	  while (outer.hasNext()) {
		E b = outer.next();
		Set<E> outSet2 = result.outLink(b);
		Set<E> intersection = new TreeSet<E>(outSet1);
		 intersection.retainAll(outSet2);
		Iterator <E> inter = intersection.iterator();
		while (inter.hasNext()) {
		    E c = inter.next();
		    Pair<E,E> pair = new ComparablePair(a,c);
		    result.remove(pair);
		}
	  }
	}
        return result;
    }

    /**
     * Returns the Kolmogorov quotient of this finite topological space, where
     * the relation is the Hasse diagram.
     * @param <E> space type
     * @return Kolmogorov quotient
     */    
    public FiniteSpace<E> T0Space() {
	Set<E> pointSet = KolmogorovQuotient().getPointSet();
	BinaryRelation<E> T0Relation 
	    = new BinaryRelation<E>(HasseDiagram());
	FiniteSpace<E> result = new FiniteSpace(pointSet,T0Relation);
	return result;
    }

    /**
     * Returns the Krull dimension of this finite topological space.
     * @return dimension
     */
    public int dimension() {
	int dim = 0; 
	FiniteSpace<E> space = new FiniteSpace(T0Space());
	BinaryRelation<E> R = space.getRelation();
	BinaryRelation<E> power1 = new BinaryRelation(R);
	BinaryRelation<E> power2 
	    = new BinaryRelation(RelationCollection.power(R,2));
	while(!power1.getPairs().equals(power2.getPairs())) {
	    power1 = RelationCollection.product(power1,R);
	    power2 = RelationCollection.product(power2,R);
	    dim = dim + 1;
	}
	return dim;
    }

    /**
     * Returns the minimal open neighbourhood of a point in this 
     * finite topological space. This consists of the point only
     * if the point does not belong to this space.
     * @param <E> space type
     * @param point
     * @return minimal open neighbourhood of point
     */
    public FiniteSpace<E> neighbourhood(E point) {
	Set<E> nPoints = new TreeSet();
	nPoints = getPreOrder().outLink(point);
	BinaryRelation<E> relatie = new BinaryRelation(getRelation());
	//FiniteSpace<E> result = new FiniteSpace(nPoints,relatie);
	FiniteSpace<E> result = new FiniteSpace();
	result.addAllPoint(nPoints);
	//result.addAllPair(relatie.getPairs());
	for (Pair<E,E> pair : relatie.getPairs()) {
	    if (nPoints.contains(pair.getKey()) 
		&& nPoints.contains(pair.getValue())) {
		result.addPair(pair);
	    }
	}
	return result;
    }

    /**
     * Returns the punctured minimal open neighbourhood of a point in this 
     * finite topological space. This is the empty space if the point 
     * does not belong to this space.
     * @param <E> space type
     * @param point
     * @return punctured minimal open neighbourhood of point
     */
    public FiniteSpace<E> pNeighbourhood(E point) {
	FiniteSpace<E> result = new FiniteSpace(neighbourhood(point));
	result.removePoint(point);
	return result;
    }

    /**
     * Returns the closure of a point in this finite topological space.
     * This consists of the point only if the point does not 
     * belong to this space.
     * @param <E> space type
     * @param point
     * @return closure of point
     */
    public FiniteSpace<E> closure(E point) {
	
	Set<E> nPoints = new TreeSet(getPreOrder().inLink(point));
	FiniteSpace<E> result = new FiniteSpace();
	result.addAllPoint(nPoints);
	for (Pair<E,E> pair : getRelation().getPairs()) {
	    if (nPoints.contains(pair.getKey())
		&& nPoints.contains(pair.getValue())) {
		result.addPair(pair);
	    }
	}
	return result;
    }

    /**
     * Returns the punctured closure of a point in this 
     * finite topological space. This is the empty space 
     * if the point does not belong to this space. 
     * @param <E> space type
     * @param point
     * @return punctured closure of point
     */
    
    public FiniteSpace<E> pClosure(E point) {
	
	FiniteSpace<E> result = new FiniteSpace(closure(point));
	result.removePoint(point);
	return result;
    }

    /**
     * Returns the set of minimal elements of the 
     * Kolmogorov quotient of this finite topological space.
     * @param <E> space type
     * @return vertices
     */ 
    public Set<E> KVertices() {
	Set<E> result = new TreeSet();
	FiniteSpace<E> space = new FiniteSpace(T0Space());
	Iterator <E> iter = space.getPointSet().iterator();
	while (iter.hasNext()) {
	    E v = iter.next();
	    Set<E> out = space.getRelation().outLink(v);
	    if (out.size() == 0) {
		result.add(v);
	    }
	}
	return result;
    }
   
    /**
     * Returns the set of maximal elements of the Kolmogorov quotient of a
     * finite topological space.
     * @param <E> space type
     * @return open points
     */
    public Set<E> KOpenPoints() {
	Set<E> result = new TreeSet();
	FiniteSpace<E> space = new FiniteSpace(T0Space());
	Iterator <E> iter = space.getPointSet().iterator();
	while (iter.hasNext()) {
	    E v = iter.next();
	    Set<E> in = space.getRelation().inLink(v);
	    if (in.size() == 0) {
		result.add(v);
	    }
	}
	return result;
    }
   
}

