package topology;

import java.util.*;

/**
The class PointReductions consists of some point-reduction methods for finite topological spaces which yield weakly homotopy equivalent spaces.
In particular, the Betti numbers remain unchanged.
@author Patrick Erik Bradley
@version 1.0
 */
public class PointReductions {

    /**
     * Returns true if the given point is an up beat point of the 
     * Kolmogorov quotient of a given finite topological space.
     * An up beat point is a point whose inLink in the Hasse Diagram
     * has precisely one element.
     * @param <E> space type
     * @param space
     * @param point
     * @return true iff point is an up beat point
     */ 
    public static <E extends Comparable<E>> boolean
			     isUpBeatPoint(FiniteSpace<E> space, E point) {
	boolean result = false;
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Set<E> pointSet = aSpace.getPointSet();
	if (pointSet.contains(point)) {
	    BinaryRelation<E> relation = 
		new BinaryRelation(aSpace.HasseDiagram());
	    Set<E> coveredBy = relation.inLink(point);
	    if (coveredBy.size() == 1) {
		result = true;
	    }
	}
	return result;
    }

    /**
     * Returns true if the given point is a down beat point of the Kolmogorov
     * quotient of a given finite topological space. A down beat point is a 
     * point whose outLink in the HasseDiagram has precisely one element.
     * @param <E> space type
     * @param space
     * @param point
     * @return true iff point is a down beat point
     */
    public static <E extends Comparable<E>> boolean
			     isDownBeatPoint(FiniteSpace<E> space, E point) {
	boolean result = false;
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Set<E> pointSet = aSpace.getPointSet();
	if (pointSet.contains(point)) {
	    BinaryRelation<E> relation =
		new BinaryRelation(aSpace.HasseDiagram());
	    Set<E> covers = relation.outLink(point);
	    if (covers.size() == 1) {
		result = true;
	    }
	}
	return result;
    }

    /**
     * Returns true if the Kolmogorov quotient of a finite topological space 
     * has no beat point. A beat point is a point whose inLink or outLink 
     * in the HasseDiagram has precisely one element.
     * @param <E> space type
     * @param space
     * @return true iff there are no up or down beat points
     */    
    public static <E extends Comparable<E>> boolean 
			     noBeatPoint(FiniteSpace<E> space) {
	
	boolean result = true;
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Iterator <E> iter = aSpace.getPointSet().iterator();
	while(iter.hasNext()) {
	    E e = iter.next();
	    if (isUpBeatPoint(aSpace,e) || isDownBeatPoint(aSpace,e)) {
		result = false;
		break;
	    }
	}
	return result;
    }


    /**
     * Returns the set of up beat points of the Kolmogorov quotient of 
     * a finite topological space.
     * @param <E> space type
     * @param space
     * @return set of up beat points
     */
    public static <E extends Comparable<E>> Set<E> 
			     upBeatPoints(FiniteSpace<E> space) {
	Set<E> result = new TreeSet();
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Set<E> pointSet = aSpace.getPointSet();
	Iterator<E> iter = pointSet.iterator();
	while (iter.hasNext()) {
	    E e = iter.next();
	    if (isUpBeatPoint(aSpace,e)) {
		result.add(e);
	
	    }  
	}
	return result;
    }

    /**
     * Returns the set of down beat points of the Kolmogorov quotient of
     * a finite topological space.
     * @param <E> space type
     * @param space
     * @return set of down beat points
     */
    public static <E extends Comparable<E>> Set<E> 
			     downBeatPoints(FiniteSpace<E> space) {
	Set<E> result = new TreeSet();
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Set<E> pointSet = aSpace.getPointSet();
	Iterator<E> iter = pointSet.iterator();
	while (iter.hasNext()) {
	    E e = iter.next();
	    if (isDownBeatPoint(aSpace,e)) {
		result.add(e);
	
	    }  
	}
	return result;
    }

    
   
    /**
     * Returns a 0-core of a the Kolmogorov quotient of a 
     * finite topological space, i.e. a space whose
     * beat points are iteratively removed. If the 0-core is a single point,
     * then the space is called contractible.
     * @param <E> space type
     * @param space
     * @return 0-core
     */
    public static <E extends Comparable<E>> FiniteSpace<E>
			     removeBeatPoints(FiniteSpace<E> space) {
	FiniteSpace<E> aSpace = new FiniteSpace();
	aSpace.addAllPoint(space.T0Space().getPointSet());
	aSpace.addAllPair(space.T0Space().getRelation().getPairs());
	FiniteSpace<E> result 
	    = new FiniteSpace(aSpace.getPointSet(),aSpace.getRelation());
	while(!noBeatPoint(result)){
	    Set<E> pointSet = result.getPointSet();
	    BinaryRelation<E> relation 
		= new BinaryRelation(result.getRelation());
	    Iterator <E> iter = pointSet.iterator();
	    while(iter.hasNext()) {
		E e = iter.next();
		if (isUpBeatPoint(result,e) || isDownBeatPoint(result,e)) {
		    result.removePoint(e);
		    break;
		}
	    }
	} 
	return result;
    }

    /**
     * Returns true if the given point is an up weak point of the 
     * Kolmogorov quotient of a given finite topological space.
     * An up weak point is a point whose punctured neighbourhood 
     * is contractible, i.e. can be reduced to a single point by successively 
     * removing beat points.
     * @param <E> space type
     * @param space
     * @param point
     * @return true iff point is up weak point
     */
    public static <E extends Comparable<E>> boolean 
			     isUpWeakPoint(FiniteSpace<E> space, E point) {
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	boolean result = false;
	Set<E> pointSet = aSpace.getPointSet();
	if (pointSet.contains(point)) {
	    FiniteSpace<E> F_point = aSpace.pNeighbourhood(point);
	    FiniteSpace<E> core = removeBeatPoints(F_point);
	    if (core.getPointSet().size() == 1) {
		result = true;
	    }
	}
	return result;
    }

    /**
     * Returns true if the given point is a down weak point of the 
     * Kolmogorov quotient of a given finite topological space.
     * A down weak point is a point whose punctured closure 
     * is contractible, i.e. can be reduced to a single point by successively 
     * removing beat points.
     * @param <E> space type
     * @param space
     * @param point
     * @return true iff point is down weak point
     */
    public static <E extends Comparable<E>> boolean 
			     isDownWeakPoint(FiniteSpace<E> space, E point) {
	FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	boolean result = false;
	Set<E> pointSet = aSpace.getPointSet();
	if (pointSet.contains(point)) {
	    FiniteSpace<E> F_point = aSpace.pClosure(point);
	    FiniteSpace<E> core = removeBeatPoints(F_point);
	    if (core.getPointSet().size() == 1) {
		result = true;
	    }
	}
	return result;
    }

    /**
     * Returns true if the Kolmogorov quotient of a 
     * finite topological space has no weak point. A weak point is a point
     * whose punctured neighbourhood or punctured closure is contractible,
     * i.e. can be reduced to a single point by successively removing 
     * beat points.
     * @param <E> space type
     * @param space
     * @return true iff space has no up or down weak points
     */
    public static <E extends Comparable<E>> boolean
			     noWeakPoint(FiniteSpace<E> space) {
	boolean result = true;
        FiniteSpace<E> aSpace = new FiniteSpace(space.T0Space());
	Set<E> pointSet = aSpace.getPointSet();
	Iterator<E> iter = pointSet.iterator();
	while (iter.hasNext()) {
	    E e = iter.next();
	    if (isUpWeakPoint(aSpace,e) || isDownWeakPoint(aSpace,e)) {
	    	result = false;
	    	break;
	    }
	}
	return result;
    }
    
    /**
     * Returns a 1-core of the Kolmogorov quotient of a finite topological
     * space, i.e. a space whose weak points are iteratively removed. 
     * A space having a 1-core consisting of a single point is called 
     * collapsible.
     * @param <E> space type
     * @param space
     * @return 1-core
     */
    public static <E extends Comparable<E>> FiniteSpace<E>
				 removeWeakPoints(FiniteSpace<E> space) {
	FiniteSpace<E> aSpace = new FiniteSpace(space);
	//aSpace.addAllPoint(space.T0Space().getPointSet());
	//aSpace.addAllPair(space.T0Space().getRelation().getPairs());
	    FiniteSpace<E> result 
		= new FiniteSpace(aSpace.getPointSet(),aSpace.getRelation());
	    while (!noWeakPoint(result)) {
		Set<E> pointSet = result.getPointSet();
		Iterator<E> iter = pointSet.iterator();
		while (iter.hasNext()) {
		    E e = iter.next();
		    if (isUpWeakPoint(result,e) || isDownWeakPoint(result,e)) {
			result.removePoint(e);
			break;
		    }
		}
	    }
	    return result;
	}


}
