package topology;

import topology.*;
import java.util.*;

/**
 * This class computes a spanning forest for a finite 
 * topological space using Prim's algorithm. Additionally,
 * the individual connected components are given out.
 * @param <E> space type
 * @author Patrick Erik Bradley
 * @version 1.0 May 30, 2014
 */
public class PrimMST<E extends Comparable<E>> {

    /**
     * the point set underlying the space for which a spanning forest
     * is to be computed
     * @param <E> space type
     */
    private TreeSet<E> spacePointSet = new TreeSet(); 

    /**
     * the relation of the space for which a spanning forest
     * is to be computed
     * @param <E> space type
     */
    private BinaryRelation<E> relation = new BinaryRelation<E>();

    /**
     * the relation defining the spanning forest
     * @param <E> space type
     */
    private BinaryRelation<E> mstRelation = new BinaryRelation<E>();

    /**
     * the list of point sets underlying the connected components of the space
     * @param <E> space type
     */
    private List<TreeSet<E>> ComponentList 
	= new ArrayList<TreeSet<E>>();
    
    /**
     * Creates a new PrimMST object.
     */
    public PrimMST() {
    }

    /**
     * Creates a spanning forest for the given finite space
     * @param <E> space type
     * @param space the finite space for which a spanning forest
     * is to be computed
     */
    public PrimMST(FiniteSpace<E> space) {

	spacePointSet = new TreeSet(space.getPointSet());
	TreeSet<E> pointSet = new TreeSet(space.getPointSet());
	relation = space.getRelation();
	TreeSet<E> visitedPoints = new TreeSet<E>();
	if (!pointSet.isEmpty()) {
	    visitedPoints.add(pointSet.first());
	    boolean test = false;
	    while(!visitedPoints.equals(pointSet)) {
		TreeSet<E> vPoints = new TreeSet(visitedPoints);
		for (E point : vPoints) {
		    TreeSet<E> inLink = new TreeSet(relation.inLink(point));
		    inLink.removeAll(visitedPoints);
		    TreeSet<E> outLink = new TreeSet(relation.outLink(point));
		    outLink.removeAll(visitedPoints);
		    if (!inLink.isEmpty()) {
			E point2 = inLink.first();
			visitedPoints.add(point2);
			Pair<E,E> pair = new ComparablePair(point2,point);
			mstRelation.add(pair);
			test = true;
			break;
		    } else {
			if (!outLink.isEmpty()) {
			    E point2 = outLink.first();
			    visitedPoints.add(point2);
			    Pair<E,E> pair = new ComparablePair(point,point2);
			    mstRelation.add(pair);
			    test = true;
			    break;
			} else {
			    test = false;
			}
		    }
		}
		if (test == false) {
		    ComponentList.add(new TreeSet(visitedPoints));
		    pointSet.removeAll(visitedPoints);
		    visitedPoints.removeAll(visitedPoints);
		    if (!pointSet.isEmpty()) {
			E point = pointSet.first();
			visitedPoints.add(point);
		    }
		}
		
	    }
	    ComponentList.add(visitedPoints);
	}
    }
    
    /**
     * Returns the binary relation defining the spanning forest.
     * @param <E> the space type
     */
    public BinaryRelation<E> getMstRelation() {
	return mstRelation;
    }

    /**
     * Returns the spanning forest as a finite space.
     * @param <E> space type
     */
    public FiniteSpace<E> getMst() {
	return new FiniteSpace(spacePointSet,mstRelation);
    }

    /**
     * Returns a list of the point sets underlying the connected components 
     * of the space.
     * @param <E> space type  
     */
    public List<TreeSet<E>> getComponentSetList() {
	return ComponentList;
    }

    /**
     * Returns a list of connected components.
     * @param <E> space type
     */
    public List<FiniteSpace<E>> getComponentList() {
	List<FiniteSpace<E>> ComponentSpaceList = new ArrayList<FiniteSpace<E>>();
	for (int i = 0; i < ComponentList.size(); i++) {
	    FiniteSpace<E> component 
		= new FiniteSpace(ComponentList.get(i),relation);
	    ComponentSpaceList.add(component);
	}
	return ComponentSpaceList;
    }

    /**
     * Returns a list of trees making up the spanning forest.
     * @param <E> space type
     */
    public List<FiniteSpace<E>> spanningTreeList() {
	List<FiniteSpace<E>> spanningTreeList = new ArrayList<FiniteSpace<E>>();
	for (int i = 0; i < ComponentList.size(); i++) {
	    FiniteSpace<E> tree 
		= new FiniteSpace(ComponentList.get(i),mstRelation);
	    spanningTreeList.add(tree);
	}
	return spanningTreeList;
    }

    /**
     * Returns the number of connected components. This is the zero-th
     * Betti number.
     */
    public int componentNumber() {
	return ComponentList.size();
    }
}
