package topology;

import java.util.*;

/**
 * The class RelationCollections consists of some methods for BinaryRelations.
 * @author Patrick Erik Bradley
 * @version 1.0
 */
public class RelationCollection {

    /**
     * Returns the product of two binary relations. 
     * @param <E> relation type
     * @param R relation 1
     * @param S relation 2
     * @return product relation
      */  
    public static <E extends Comparable<E>> BinaryRelation<E> 
	product(BinaryRelation<E> R, BinaryRelation<E> S) 
    {
	BinaryRelation<E> result = new BinaryRelation();
	Iterator<Pair<E,E>> iterR = R.getPairs().iterator();
	while (iterR.hasNext()) {
		Pair<E,E> currR = (Pair<E,E>)iterR.next(); 
		Iterator<Pair<E,E>> iterS = S.getPairs().iterator();
		while (iterS.hasNext()) {
		    Pair<E,E> currS = (Pair<E,E>)iterS.next();
		    if (currR.getValue().equals(currS.getKey())) {
			Pair<E,E> p = new 
			    ComparablePair(currR.getKey(),currS.getValue());
			result.add(p);
		    }
		}
	}
	return result;
    }

    /**
     * Returns the n-th power of a binary relation. 
     * This is the empty relation if n is negative.
     * @param <E> relation type
     * @param R a relation
     * @param n the exponent
     * @return n-th power of relation
     */
    public static <E extends Comparable<E>> BinaryRelation<E>
			     power(BinaryRelation<E> R, int n) {
	BinaryRelation<E> result = new BinaryRelation();
	if (n == 0) {
	    Iterator<Pair<E,E>> iter = R.getPairs().iterator();
	    while (iter.hasNext()) {
		Pair<E,E> pair = (Pair<E,E>)iter.next();
		E key = pair.getKey();
		E value = pair.getValue();
		result.add(new ComparablePair(key,key));
		result.add(new ComparablePair(value,value));
	    }
	}
	if (n>0) {
	    result.addAll(R.getPairs());
	    for (int i=1; i< n+1; i++){
		result = product(result,R);
	    }
	}
	return result;
    }

    /**
     * Returns the transitive closure of a relation.
     * @param <E> relation type
     * @param R a relation
     * @return transitive closure of relation
     */
    public static <E extends Comparable<E>> BinaryRelation<E>
			     transitiveClosure(BinaryRelation<E> R) {
	BinaryRelation<E> power1 = new BinaryRelation(R.getPairs());
	BinaryRelation<E> power2 = new BinaryRelation(power(R,2).getPairs());
	BinaryRelation<E> result = new BinaryRelation(R.getPairs());
	while(!power1.getPairs().equals(power2.getPairs())) {
	    power1 = product(power1,R);
	    power2 = product(power2,R);
	    result.addAll(power1.getPairs());
	}
	return result;
    }
    
}
