package topology;

import java.util.*;

/**
Some examples of finite topological spaces. 
@author Patrick Erik Bradley
@version 1.0
 */

public class TopologyExamples {
    
    /**
     * Returns the empty space. All its Betti numbers vanish.
     * @param <E> space type
     * @return empty space
     */
    public static <E extends Comparable<E>> FiniteSpace<E> emptySpace() {
	FiniteSpace<E> result = new FiniteSpace();
	return result;
    }
    
    /**
     * Returns a parametrised singleton space. Its zero-th Betti number is one,
     * all other Betti numbers vanish.
     * @param <E> space type
     * @param e a point
     * @return one-point space
     */
    public static <E extends Comparable<E>> FiniteSpace<E> pointSpace(E e) {
	FiniteSpace<E> result = new FiniteSpace();
	result.addPoint(e);
	return result;
    }

    /**
     * Returns a circle. Its zero-th Betti number is zero, 
     * and its first Betti number is one.
     * @return minimal circle
     */
    public static FiniteSpace circle() {
	FiniteSpace<Integer> circle = new FiniteSpace();
	Set<Integer> circlePointSet = new TreeSet();
	circlePointSet.addAll(Arrays.asList(1,2,3,4));
	BinaryRelation<Integer> circleRelation = new BinaryRelation();
	circleRelation.add(new ComparablePair(1,3));
	circleRelation.add(new ComparablePair(1,4));
	circleRelation.add(new ComparablePair(2,3));
	circleRelation.add(new ComparablePair(2,4));
	circle = new FiniteSpace(circlePointSet,circleRelation);
	return circle;
    }

    /**
     * Returns a 2-sphere. Its zero-th Betti number is one, 
     * its first Betti number is zero, and its second
     * Betti number is one.
     * @return minimal 2-sphere
     */
    public static FiniteSpace twoSphere() {
	FiniteSpace<Integer> twoSphere = new FiniteSpace();
	FiniteSpace<Integer> twoPoints = new FiniteSpace();
	twoPoints.addPoint(5);
	twoPoints.addPoint(6);
	twoSphere = TopoCollection.nonHausdorffJoin(circle(),twoPoints);
	return twoSphere;
    }

    /**
     * Returns the wallet. It is a finite space which is not contractible,
     * but can be reduced to a point by iteratively removing weak points.
     * Hence, its zero-th Betti number is one, 
     * and its higher Betti numbers vanish.
     * @return wallet
     */
    public static FiniteSpace wallet() {
	FiniteSpace<Integer> wallet = new FiniteSpace();
	Set<Integer> walletPointSet = new TreeSet();
	walletPointSet.addAll(Arrays.asList(1,2,3,4,5,6,7,8,9,10,11));
	BinaryRelation<Integer> walletRelation = new BinaryRelation();
	walletRelation.add(new ComparablePair(1,5));
	walletRelation.add(new ComparablePair(1,6));
	walletRelation.add(new ComparablePair(2,5));
	walletRelation.add(new ComparablePair(2,7));
	walletRelation.add(new ComparablePair(3,6));
	walletRelation.add(new ComparablePair(3,8));
	walletRelation.add(new ComparablePair(4,7));
	walletRelation.add(new ComparablePair(4,8));
	walletRelation.add(new ComparablePair(5,9));
	walletRelation.add(new ComparablePair(5,10));
	walletRelation.add(new ComparablePair(6,9));
	walletRelation.add(new ComparablePair(6,10));
	walletRelation.add(new ComparablePair(7,10));
	walletRelation.add(new ComparablePair(7,11));
	walletRelation.add(new ComparablePair(8,10));
	walletRelation.add(new ComparablePair(8,11));
	wallet = new FiniteSpace(walletPointSet,walletRelation);
	return wallet;
    }

    /**
     * Returns some space. It consists of a fat point with three elements
     * above a chain of two points: F -> 4 -> 5. Its Betti numbers are the 
     * same as that of the singleton space.
     * @return some space
     */
    public static FiniteSpace someSpace() {
	FiniteSpace<Integer> result = new FiniteSpace();
	Set<Integer> somePointSet = new TreeSet();
	somePointSet.addAll(Arrays.asList(1,2,3,4,5));
	BinaryRelation<Integer> someRelation = new BinaryRelation();
	someRelation.add(new ComparablePair(1,2));
	someRelation.add(new ComparablePair(2,3));
	someRelation.add(new ComparablePair(3,1));
	someRelation.add(new ComparablePair(3,4));
	someRelation.add(new ComparablePair(4,5));
	result = new FiniteSpace(somePointSet,someRelation);
	return result;
    }
    
    /** 
     * Returns my space. It is the incidence space of a full triangle
     * and a triangle without interior face having one common side.
     * Its zero-th and first Betti numbers are one, all higher Betti numbers
     * vanish.
     * @return my space     
     */
    public static FiniteSpace mySpace() {
	FiniteSpace<String> result = new FiniteSpace();
	Set<String> myPointSet = new TreeSet<String>();
	BinaryRelation<String> myRelation = new BinaryRelation<String>();
	Set<String> emptySet = new TreeSet<String>();
	Set<String> vertices = new TreeSet<String>();
	Set<String> lines = new TreeSet<String>();
	Set<String> areas = new TreeSet<String>();
	vertices.addAll(Arrays.asList("0","1","2","3"));
	lines.addAll(Arrays.asList("01","02","12","13","23"));
	areas.add("123");

	myPointSet.addAll(vertices);
	myPointSet.addAll(lines);
	myPointSet.addAll(areas);

	myRelation.add(new ComparablePair("123","12"));
	myRelation.add(new ComparablePair("123","13"));
	myRelation.add(new ComparablePair("123","23"));
	myRelation.add(new ComparablePair("12","1"));
	myRelation.add(new ComparablePair("12","2"));
	myRelation.add(new ComparablePair("13","1"));
	myRelation.add(new ComparablePair("13","3"));
	myRelation.add(new ComparablePair("23","2"));
	myRelation.add(new ComparablePair("23","3"));
	myRelation.add(new ComparablePair("01","0"));
	myRelation.add(new ComparablePair("01","1"));
	myRelation.add(new ComparablePair("02","0"));
	myRelation.add(new ComparablePair("02","2"));

	result = new FiniteSpace(myPointSet,myRelation);
	return result;
    }
    
    /**
     * Returns a random finite space with n points.
     * @param n number of points
     * @return random space
     */
    public static FiniteSpace randomSpace(int n) {
	
	Random randomGenerator = new Random();
	Set<Integer> pointSet = new TreeSet();
	BinaryRelation<Integer> relation = new BinaryRelation();
	for (int i = 0; i < n; i++) {
	    pointSet.add(i);
	}
	for (int i = 0; i < n; i++) {
	    for (int j = 0; j < n; j++) {
		int randomInt = randomGenerator.nextInt(2);
		if (randomInt == 1) {
		    relation.add(new ComparablePair(i,j));
		}
	    }
	}
	FiniteSpace<Integer> result = new FiniteSpace(pointSet,relation);
	return result;
    }

    /**
     * Returns yet another finite space.
     */
    public static FiniteSpace yas() {
	FiniteSpace<String> yas = new FiniteSpace();
	Set<String> yaPointSet = new TreeSet();
	Set<String> vertices = new TreeSet();
	vertices.addAll(Arrays.asList("1","2","3","4","5","6"));
	Set<String> lines = new TreeSet();
	lines.addAll(Arrays.asList("a","b","c","d","e","f"));
	Set<String> faces = new TreeSet();
	faces.addAll(Arrays.asList("A","B"));
	yaPointSet.addAll(vertices);
	yaPointSet.addAll(lines);
	yaPointSet.addAll(faces);
	BinaryRelation<String> yaRelation = new BinaryRelation();
	yaRelation.add(new ComparablePair("A","a"));
	yaRelation.add(new ComparablePair("A","b"));
	yaRelation.add(new ComparablePair("A","c"));
	yaRelation.add(new ComparablePair("A","d"));
	yaRelation.add(new ComparablePair("A","e"));
	yaRelation.add(new ComparablePair("A","f"));
	yaRelation.add(new ComparablePair("B","d"));
	yaRelation.add(new ComparablePair("B","e"));
	yaRelation.add(new ComparablePair("B","f"));
	yaRelation.add(new ComparablePair("a","1"));
	yaRelation.add(new ComparablePair("a","2"));
	yaRelation.add(new ComparablePair("b","2"));
	yaRelation.add(new ComparablePair("b","3"));
	yaRelation.add(new ComparablePair("c","1"));
	yaRelation.add(new ComparablePair("c","3"));
	yaRelation.add(new ComparablePair("d","4"));
	yaRelation.add(new ComparablePair("d","5"));
	yaRelation.add(new ComparablePair("e","5"));
	yaRelation.add(new ComparablePair("e","6"));
	yaRelation.add(new ComparablePair("f","4"));
	yaRelation.add(new ComparablePair("f","6"));
	yas = new FiniteSpace(yaPointSet,yaRelation);
	return yas;

    }
}
